DECLARE SUB TransformPts (M() AS DOUBLE, cam AS ANY, target AS ANY)
DECLARE SUB Lookat (CameraMat() AS DOUBLE, target AS ANY, CameraPos AS ANY, rollangle!)
DECLARE SUB Perspective ()
DECLARE SUB Scanconvert ()
DECLARE SUB SetIdentity (Mat() AS DOUBLE)
DECLARE SUB read3dg (filename$)
DECLARE SUB PointCameraWithRoll (CameraMat() AS DOUBLE, Fwd AS ANY, rollangle!)
DECLARE SUB PrintVector (v AS ANY)
DECLARE SUB PrintMatrix (Mat() AS DOUBLE)
DECLARE SUB PointCamera (M() AS DOUBLE, Fwd AS ANY, Up0 AS ANY)
DECLARE SUB cross (V1 AS ANY, V2 AS ANY, Vout AS ANY)
DECLARE FUNCTION dot! (v AS ANY, w AS ANY)
DECLARE SUB normalize (v AS ANY)
DECLARE FUNCTION magnitude! (v AS ANY)

CONST pi = 3.14159265358979#
CONST deg2rad = pi / 180

'make these larger to handle larger models
CONST maxpts = 30, maxtris = 50

COMMON SHARED numverts%, numtris%

TYPE vector
	x AS DOUBLE
	y AS DOUBLE
	z AS DOUBLE
END TYPE
TYPE vec2d
	x AS INTEGER
	y AS INTEGER
END TYPE
DIM SHARED modelpts(maxpts) AS vector 'model-space points
DIM SHARED rotpts(maxpts) AS vector   'camera transformed points
DIM SHARED screenpts(maxpts) AS vec2d 'screen-space points
DIM SHARED oldpts(maxpts) AS vec2d 'screen-space points
DIM SHARED trinorm(maxtris) AS vector 'triangle face normals
DIM SHARED tri%(maxtris, 0 TO 2) 'contains a list of indexes to vertices of each triangle
DIM CameraMat(0 TO 2, 0 TO 2) AS DOUBLE
CALL SetIdentity(CameraMat())

DIM Fwd AS vector: Fwd.x = 0: Fwd.y = 0: Fwd.z = 1
DIM Up0 AS vector: Up0.x = 0: Up0.y = 1: Up0.z = 0
DIM cam AS vector: cam.x = 0: cam.y = 80: cam.z = -170
DIM target AS vector: target.x = 0: target.y = 54: target.z = 80

'-------------------main----------------------
CLS
CALL read3dg("house.3dg")
LOCATE 10, 1: PRINT "# of vertices="; numverts%
PRINT "# of triangles="; numtris%

SCREEN 12
PRINT "Camera Lookat Demo by Toshi July 07th to 10th, 2001";
COLOR 14
PRINT "[5]/[+]=zoom in, [0]/[-]=zoom out"
PRINT "use number pad to navigate 8=forward, 2=back, 4=left, 6=right, 7=up, 1=down "
COLOR 15
DO
'CALL PointCamera(CameraMat(), Fwd, Up0)
'CALL PointCameraWithRoll(CameraMat(), Fwd, rollangle)
CALL Lookat(CameraMat(), cam, target, 0)

'CALL SetIdentity(CameraMat())
'CALL PrintMatrix(CameraMat())

CALL TransformPts(CameraMat(), cam, target)

CALL Perspective
CALL Scanconvert
k$ = INKEY$
SELECT CASE k$
	CASE "6"
	   cam.x = cam.x + 10
	CASE "4"
	   cam.x = cam.x - 10
	CASE "1"
	   cam.y = cam.y - 10
	CASE "7"
	   cam.y = cam.y + 10
	CASE "8"
	   cam.z = cam.z + 10
	CASE "2"
	   cam.z = cam.z - 10
	CASE "5", "+"
	   cam.x = cam.x + (target.x - cam.x) * .1
	   cam.y = cam.y + (target.y - cam.y) * .1
	   cam.z = cam.z + (target.z - cam.z) * .1
	CASE "0", "-"
	   cam.x = cam.x - (target.x - cam.x) * .1
	   cam.y = cam.y - (target.y - cam.y) * .1
	   cam.z = cam.z - (target.z - cam.z) * .1
	CASE CHR$(27)
		EXIT DO
	CASE ELSE
END SELECT

LOOP
END
10 RESUME NEXT

SUB cross (V1 AS vector, V2 AS vector, Vout AS vector)
'vector Vout = V1 x V2
Vout.x = V1.y * V2.z - V2.y * V1.z
Vout.y = V2.x * V1.z - V1.x * V2.z
Vout.z = V1.x * V2.y - V2.x * V1.y
END SUB

FUNCTION dot (v AS vector, w AS vector)
dot = v.x * w.x + v.y * w.y + v.z * w.z
END FUNCTION

SUB Lookat (CameraMat() AS DOUBLE, target AS vector, CameraPos AS vector, rollangle)
DIM Fwd AS vector
Fwd.x = target.x - CameraPos.x
Fwd.y = target.y - CameraPos.y
Fwd.z = target.z - CameraPos.z

'may not be needed
CALL normalize(Fwd)

CALL PointCameraWithRoll(CameraMat(), Fwd, rollangle)

END SUB

FUNCTION magnitude (v AS vector)
magnitude = SQR(v.x * v.x + v.y * v.y + v.z * v.z)
END FUNCTION

SUB matmul (m1() AS SINGLE, m2() AS SINGLE, m3() AS SINGLE)
'm3 = m1*m2
'm1 and m2 are 3x3 matrices
m3(0, 0) = m1(0, 0) * m2(0, 0) + m1(0, 1) * m2(1, 0) + m1(0, 2) * m2(2, 0)
m3(1, 0) = m1(1, 0) * m2(0, 0) + m1(1, 1) * m2(1, 0) + m1(1, 2) * m2(2, 0)
m3(2, 0) = m1(2, 0) * m2(0, 0) + m1(2, 1) * m2(1, 0) + m1(2, 2) * m2(2, 0)

m3(0, 1) = m1(0, 0) * m2(0, 1) + m1(0, 1) * m2(1, 1) + m1(0, 2) * m2(2, 1)
m3(1, 1) = m1(1, 0) * m2(0, 1) + m1(1, 1) * m2(1, 1) + m1(1, 2) * m2(2, 1)
m3(2, 1) = m1(2, 0) * m2(0, 1) + m1(2, 1) * m2(1, 1) + m1(2, 2) * m2(2, 1)

m3(0, 2) = m1(0, 0) * m2(0, 2) + m1(0, 1) * m2(1, 2) + m1(0, 2) * m2(2, 2)
m3(1, 2) = m1(1, 0) * m2(0, 2) + m1(1, 1) * m2(1, 2) + m1(1, 2) * m2(2, 2)
m3(2, 2) = m1(2, 0) * m2(0, 2) + m1(2, 1) * m2(1, 2) + m1(2, 2) * m2(2, 2)
END SUB

SUB normalize (v AS vector)
M = v.x * v.x + v.y * v.y + v.z * v.z
IF M <> 0 THEN
  invmag = 1 / SQR(M)
  v.x = v.x * invmag
  v.y = v.y * invmag
  v.z = v.z * invmag
ELSE
  'BEEP
  v.x = 1
  v.y = 0
  v.z = 0
END IF
END SUB

SUB Perspective
FOR i% = 0 TO numverts% - 1
	'IF frontfacing(i%) THEN
		rz = rotpts(i%).z  'fix this later to move camera to PRP
		'IF rz <= 0 THEN BEEP
		screenpts(i%).x = 320 + CINT(rotpts(i%).x * 200 / rz)
		screenpts(i%).y = 240 + CINT(rotpts(i%).y * 200 / rz)
		'PRINT ">"; rotpts(i%).x; ","; rotpts(i%).y; ","; rotpts(i%).z
		'PRINT i%; ":"; screenpts(i%).x; ","; screenpts(i%).y
		'COLOR 14, 1: CIRCLE (screenpts(i%).x, screenpts%(i%).y), i%, 14
	'END IF
NEXT i%
END SUB

SUB PointCamera (M() AS DOUBLE, Fwd AS vector, Up0 AS vector)
'one form of the lookat transform
'PRE:
'  Fwd is the foward direction vector of the camera
'  Up0 is the up vector that determines the roll of camera
'POST:
'  M is the set to contain the appropriate camera matrix

DIM Fd AS vector
Fd = Fwd

DIM Up AS vector
Up = Up0

'normalize the forward vector
CALL normalize(Fd)

t = dot(Fd, Up)

'The new Up = Up0 aligned perpendicular to forward vector.
Up.x = Up.x - t * Fd.x
Up.y = Up.y - t * Fd.y
Up.z = Up.z - t * Fd.z

'normalize the Up vector
CALL normalize(Up)

'the right-direction vector
DIM Rt AS vector

'Rt = Up x Fd
CALL cross(Up, Fd, Rt)

'set up the camera matrix using the derived basis vectors
M(0, 0) = Rt.x: M(0, 1) = Rt.y: M(0, 2) = Rt.z
M(1, 0) = Up.x: M(1, 1) = Up.y: M(1, 2) = Up.z
M(2, 0) = Fd.x: M(2, 1) = Fd.y: M(2, 2) = Fd.z



END SUB

SUB PointCameraWithRoll (CameraMat() AS DOUBLE, Fwd AS vector, rollangle)
'PRE: pi must be defined as 3.14159...
'     deg2rad must be defined as pi/180
'     rollangle is the left to right roll angle (in degrees) of the camera
'        normally, it's -45 to 45 degrees from vertical.
angle = pi / 2 - rollangle * deg2rad
DIM Up0 AS vector
Up0.x = COS(angle)
Up0.y = SIN(angle)
Up0.z = 0
CALL PointCamera(CameraMat(), Fwd, Up0)
END SUB

SUB PrintMatrix (Mat() AS DOUBLE)
FOR r = 0 TO 2
PRINT "[";
FOR c = 0 TO 2
PRINT USING "+##.##"; Mat(r, c);
NEXT
PRINT "]"
NEXT
END SUB

SUB PrintVector (v AS vector)
PRINT USING "<###.##,"; v.x;
PRINT USING "###.##,"; v.y;
PRINT USING "###.##>"; v.z
END SUB

SUB read3dg (filename$)
a$ = "x"
OPEN filename$ FOR INPUT AS #1
i% = 0
j% = 0
ON ERROR GOTO 10
DO
INPUT #1, a%
SELECT CASE a%
	CASE 0
		'PRINT "comment"
		LINE INPUT #1, a$
	CASE 1
		INPUT #1, modelpts(i%).x, modelpts(i%).y, modelpts(i%).z
		'PRINT "v"; modelpts(i%).x, modelpts(i%).y, modelpts(i%).z
		modelpts(i%).z = modelpts(i%).z
		i% = i% + 1
		numverts% = i%
	CASE 2
		INPUT #1, tri%(j%, 0), tri%(j%, 1), tri%(j%, 2)
		'PRINT "t"; tri%(j%, 0), tri%(j%, 1), tri%(i%, 2)
		j% = j% + 1
		numtris% = j%
	CASE ELSE
		PRINT "unknown"
	END SELECT
LOOP UNTIL EOF(1)
CLOSE #1
DIM dv1 AS vector
DIM dv2 AS vector
FOR k% = 0 TO numtris% - 1
		vcx = modelpts(tri%(k%, 1)).x
		vcy = modelpts(tri%(k%, 1)).y
		vcz = modelpts(tri%(k%, 1)).z
		dv1.x = modelpts(tri%(k%, 0)).x - vcx
		dv1.y = modelpts(tri%(k%, 0)).x - vcy
		dv1.z = modelpts(tri%(k%, 0)).x - vcz
		dv2.x = modelpts(tri%(k%, 2)).x - vcx
		dv2.y = modelpts(tri%(k%, 2)).x - vcy
		dv2.z = modelpts(tri%(k%, 2)).x - vcz
		CALL cross(dv1, dv2, trinorm(k%))
NEXT k%

END SUB

SUB Scanconvert
WAIT &H3DA, 8

'erase
FOR i% = 0 TO numtris% - 1
	idx0% = tri%(i%, 0)
	idx1% = tri%(i%, 1)
	idx2% = tri%(i%, 2)
	f0% = (oldpts(idx0%).x <> screenpts(idx0%).x) OR (oldpts(idx0%).y <> screenpts(idx0%).y)
	f1% = (oldpts(idx1%).x <> screenpts(idx1%).x) OR (oldpts(idx1%).y <> screenpts(idx1%).y)
	f2% = (oldpts(idx2%).x <> screenpts(idx2%).x) OR (oldpts(idx2%).y <> screenpts(idx2%).y)
	IF f0% OR f1% THEN
		LINE (oldpts(idx0%).x, oldpts(idx0%).y)-(oldpts(idx1%).x, oldpts(idx1%).y), 0
	END IF
	IF f1% OR f2% THEN
		LINE (oldpts(idx1%).x, oldpts(idx1%).y)-(oldpts(idx2%).x, oldpts(idx2%).y), 0
	END IF
	IF f0% OR f2% THEN
		LINE (oldpts(idx0%).x, oldpts(idx0%).y)-(oldpts(idx2%).x, oldpts(idx2%).y), 0
	END IF
NEXT i%

'draw
FOR i% = 0 TO numtris% - 1
	idx0% = tri%(i%, 0)
	idx1% = tri%(i%, 1)
	idx2% = tri%(i%, 2)
	'PRINT idx0%; ","; idx1%; ","; idx2%

	LINE (screenpts(idx0%).x, screenpts(idx0%).y)-(screenpts(idx1%).x, screenpts(idx1%).y)
	LINE (screenpts(idx1%).x, screenpts(idx1%).y)-(screenpts(idx2%).x, screenpts(idx2%).y)
	LINE (screenpts(idx0%).x, screenpts(idx0%).y)-(screenpts(idx2%).x, screenpts(idx2%).y)
NEXT i%

FOR i% = 0 TO numtris% - 1
	oldpts(i%) = screenpts(i%)
NEXT
END SUB

SUB SetIdentity (Mat() AS DOUBLE)
Mat(0, 0) = 1: Mat(0, 1) = 0: Mat(0, 2) = 0
Mat(1, 0) = 0: Mat(1, 1) = 1: Mat(1, 2) = 0
Mat(2, 0) = 0: Mat(2, 1) = 0: Mat(2, 2) = 1
END SUB

SUB TransformPts (M() AS DOUBLE, cam AS vector, target AS vector)
vrp.x = cam.x
vrp.y = cam.y
vrp.z = cam.z


FOR i% = 0 TO numverts% - 1
	xo = modelpts(i%).x - vrp.x
	yo = modelpts(i%).y - vrp.y
	zo = modelpts(i%).z - vrp.z

	rotpts(i%).x = M(0, 0) * xo + M(0, 1) * yo + M(0, 2) * zo
	rotpts(i%).y = M(1, 0) * xo + M(1, 1) * yo + M(1, 2) * zo
	rotpts(i%).z = M(2, 0) * xo + M(2, 1) * yo + M(2, 2) * zo
	'PRINT rotpts(i%).x; ","; rotpts(i%).y; ","; rotpts(i%).z
NEXT i%
END SUB

